# coding: utf-8
from functools import update_wrapper

from django.urls import reverse
from django.conf import settings
from django.contrib.admin.sites import AdminSite
from django_otp.admin import OTPAdminSite

from concrete_datastore.admin.admin_form import (
    MyAuthForm,
    OTPAuthenticationForm,
)

EXTRA_MODEL_OBJECT_NAMES = ('Email', 'EmailDevice')
AUTH_MODEL_OBJECT_NAMES = (
    'User',
    'Group',
    'AuthToken',
    'ConcretePermission',
    'ConcreteRole',
)


def get_admin_site():
    if settings.USE_TWO_FACTOR_AUTH:
        admin_site = OTPAdminSite
        admin_site_kwargs = {'name': 'admin'}
        login_template = "admin/mfa-login.html"
        login_form = OTPAuthenticationForm
    else:
        admin_site = AdminSite
        admin_site_kwargs = {}
        login_template = "admin/login.html"
        login_form = MyAuthForm

    class CustomAdminSite(admin_site):
        def get_urls(self):
            from django.urls import include, path, re_path

            # Since this module gets imported in the application's root package,
            # it cannot import models from other applications at the module level,
            # and django.contrib.contenttypes.views imports ContentType.
            from django.contrib.contenttypes import views as contenttype_views

            def wrap(view, cacheable=False):
                def wrapper(*args, **kwargs):
                    return self.admin_view(view, cacheable)(*args, **kwargs)

                wrapper.admin_site = self
                return update_wrapper(wrapper, view)

            # Admin-site-wide views.
            urlpatterns = [
                path('', wrap(self.index), name='index'),
                path('login/', self.login, name='login'),
                path('logout/', wrap(self.logout), name='logout'),
                path(
                    'password_change/',
                    wrap(self.password_change, cacheable=True),
                    name='password_change',
                ),
                path(
                    'password_change/done/',
                    wrap(self.password_change_done, cacheable=True),
                    name='password_change_done',
                ),
                path(
                    'jsi18n/',
                    wrap(self.i18n_javascript, cacheable=True),
                    name='jsi18n',
                ),
                path(
                    'r/<int:content_type_id>/<path:object_id>/',
                    wrap(contenttype_views.shortcut),
                    name='view_on_site',
                ),
            ]

            # Add in each model's views, and create a list of valid URLS for the
            # app_index
            valid_app_labels = []
            for model, model_admin in self._registry.items():
                urlpatterns += [
                    path(
                        '%s/' % (model._meta.model_name),
                        include(model_admin.urls),
                    )
                ]
                if model._meta.app_label not in valid_app_labels:
                    valid_app_labels.append(model._meta.app_label)

            # If there were ModelAdmins registered, we should have a list of app
            # labels for which we need to allow access to the app_index view,
            if valid_app_labels:
                regex = r'^(?P<app_label>' + '|'.join(valid_app_labels) + ')/$'
                urlpatterns += [
                    re_path(regex, wrap(self.app_index), name='app_list')
                ]

            return urlpatterns

        def get_app_list(self, request):
            app_dict = self._build_app_dict(request)
            if not app_dict:
                #: This case should never occur, due to the admin site
                #: permssions set to "superuser".
                return []  # pragma: no cover

            # Sort the apps alphabetically.
            app_list = sorted(
                app_dict.values(), key=lambda x: x['name'].lower()
            )

            # Sort the models alphabetically within each app.
            for app in app_list:
                app['models'].sort(key=lambda x: x['name'])
            concrete_app_filter = filter(
                lambda x: x['app_label'].lower() == 'concrete', app_list
            )
            concrete_app = next(concrete_app_filter)
            models = concrete_app.pop('models')
            model_group_models = {
                'group_name': 'Models',
                **concrete_app,
                'models': [
                    model
                    for model in models
                    if model['object_name']
                    not in EXTRA_MODEL_OBJECT_NAMES + AUTH_MODEL_OBJECT_NAMES
                ],
            }
            extra_group_models = {
                'group_name': 'Extra',
                **concrete_app,
                'models': [
                    model
                    for model in models
                    if model['object_name'] in EXTRA_MODEL_OBJECT_NAMES
                ],
            }
            auth_group_models = {
                'group_name': 'Auth',
                **concrete_app,
                'models': [
                    model
                    for model in models
                    if model['object_name'] in AUTH_MODEL_OBJECT_NAMES
                ],
            }
            custom_app_groups_list = [
                auth_group_models,
                extra_group_models,
                model_group_models,
            ]
            return custom_app_groups_list

        def index(self, request, extra_context=None, *args, **kwargs):
            if extra_context is None:
                extra_context = {}
            extra_context[
                'display_datamodel'
            ] = settings.ENABLE_SERVE_DATAMODEL
            if settings.USE_CORE_AUTOMATION:
                extra_context.update(
                    {
                        'use_core_automation': True,
                        'target_admin_view': reverse('coreAdmin:index'),
                        'target_admin_view_name': 'Core Admin',
                    }
                )
            return super().index(request, extra_context, *args, **kwargs)

    CustomAdminSite.login_template = login_template
    CustomAdminSite.login_form = login_form

    return CustomAdminSite(
        **admin_site_kwargs  # pylint: disable = unexpected-keyword-arg
    )



# coding: utf-8
import json

from django.core.management.base import BaseCommand, CommandError

from rest_framework import renderers

from concrete_datastore.interfaces.openapi_schema_generator import (
    SchemaGenerator,
    InvalidTokenUser,
)
from concrete_datastore.routes.urls import (
    urlpatterns as full_urls,
    api_v1_urls,
    api_v1_1_urls,
)

ALLOWED_USER_LEVELS = [
    'superuser',
    'admin',
    'manager',
    'simpleuser',
    'blocked',
]


class Command(BaseCommand):
    help = "Generates configured API schema for project."

    def add_arguments(self, parser):
        parser.add_argument(
            '--format',
            dest="format",
            choices=['yaml', 'json'],
            default='yaml',
            type=str,
            help='Select the OpenAPI Specification format',
        )
        parser.add_argument(
            '--servers',
            dest="servers",
            default=None,
            type=str,
            help=(
                'List of servers on which the server is running.'
                ' Example: ["http://localhost:8000/"]'
            ),
        )
        parser.add_argument(
            '--token',
            dest="user_token",
            default=None,
            type=str,
            help='Token of the user requesting the OpenAPI Specification',
        )
        parser.add_argument(
            '--level',
            dest="user_level",
            choices=ALLOWED_USER_LEVELS,
            default=None,
            type=str,
            help='Level of the user requesting the OpenAPI Specification',
        )
        parser.add_argument(
            '--api_version',
            dest="api_version",
            choices=['v1.1', 'v1', 'all'],
            default='v1.1',
            type=str,
            help='Select the API version for the OpenAPI Specification',
        )

    def handle(self, *args, **options):
        url_patterns_dict = {
            'v1': (api_v1_urls,),
            'v1.1': (api_v1_1_urls,),
            'all': full_urls,
        }
        servers = options['servers']
        if isinstance(servers, str):
            servers = json.loads(servers)

        user_token = options['user_token']

        if user_token is not None and options['user_level'] is not None:
            raise CommandError(
                'You should either select a token or a level, not both'
            )

        #:  if neither user_level not user_token were given
        #:  the default user should be an AnonymousUser which
        #:  matches the permissions of a `blocked` user
        user_level = options['user_level'] or 'blocked'

        #:  These verifications are added so that the unittests pass.
        #:  The `call_command` method does not check if the selected
        #:  option is in the choices list
        if options['format'] not in ['yaml', 'json']:
            raise CommandError('format should be one of yaml, json')
        if options['api_version'] not in ['v1.1', 'v1', 'all']:
            raise CommandError('format should be one of v1.1, v1, all')
        if user_level.lower().strip() not in ALLOWED_USER_LEVELS:
            raise CommandError(
                'level should be one of {}'.format(
                    ', '.join(ALLOWED_USER_LEVELS)
                )
            )

        from_db = user_token is not None

        try:
            generator = SchemaGenerator(
                servers=servers,
                patterns=url_patterns_dict[options['api_version']],
                user_token=user_token,
                user_level=user_level,
                from_db=from_db,
            )
        except InvalidTokenUser:
            raise CommandError(f'Invalid user token {user_token}.')
        schema = generator.get_schema(request=None, public=True)
        renderer = self.get_renderer(spec_format=options['format'])
        output = renderer.render(schema, renderer_context={})
        self.stdout.write(output.decode())

    def get_renderer(self, spec_format):
        renderer_cls = {
            'yaml': renderers.OpenAPIRenderer,
            'json': renderers.JSONOpenAPIRenderer,
        }[spec_format]
        return renderer_cls()



# coding: utf-8
import logging
from importlib import import_module
from tenacity import (
    Retrying,
    wait_fixed,
    retry_if_exception_type,
    stop_after_attempt,
)

from django.conf import settings

from concrete_datastore.settings.celery import app
from concrete_mailer.preparers import prepare_email

from concrete_datastore.concrete.models import Email  # pylint:disable=E0611

logger = logging.getLogger(__name__)


@app.task
def async_run_plugin_tasks():
    if isinstance(settings.PLUGINS_TASKS_FUNC, dict) is False:
        return
    for path_task, queue in settings.PLUGINS_TASKS_FUNC.items():
        module_name, func_name = path_task.rsplit('.', 1)
        module = import_module(module_name)
        function = getattr(module, func_name)
        function.apply_async(queue=queue)


@app.task
def send_async_mails(email_pk):
    email = Email.objects.get(pk=email_pk)
    perform_send_email(instance=email, is_async=True)


def save_instance_if_async(instance, is_async):
    #: If this method is callled from an async process, save the instance
    #: Otherwise, this method is called directly from the pre_save
    #: signal and the instance will be saved after the signal
    if is_async is True:
        instance.save()


def perform_send_email(instance, enable_retrying=True, is_async=False):
    def _perform_send_email():
        #: Refresh instance
        if not instance.subject or not instance.body or not instance.receiver:
            instance.resource_status = 'send-error'
            instance.resource_message = 'Some fields are empty'
            save_instance_if_async(instance=instance, is_async=is_async)
            return

        if instance.created_by is None:
            instance.resource_status = 'send-error'
            instance.resource_message = 'No sender available'
            save_instance_if_async(instance=instance, is_async=is_async)
            return
        use_tls = not getattr(settings, 'DEBUG', False)
        email = prepare_email(
            context={"sender_name": settings.EMAIL_SENDER_NAME},
            css=settings.EMAIL_CSS,
            html=instance.body,
            title=instance.subject,
            sender=settings.SERVER_EMAIL,
            recipients=[instance.receiver.email],
            use_tls=use_tls,
            email_host=settings.EMAIL_HOST,
            email_port=settings.EMAIL_PORT,
            email_host_user=settings.EMAIL_HOST_USER,
            email_host_password=settings.EMAIL_HOST_PASSWORD,
        )
        email.send()

        instance.resource_status = 'sent'
        instance.resource_message = 'Mail successfuly sent'
        save_instance_if_async(instance=instance, is_async=is_async)

    try:
        if enable_retrying is False:
            _perform_send_email()
        else:
            send_email_retryer = Retrying(
                retry=retry_if_exception_type(),
                wait=wait_fixed(settings.RETRYING_DELAY_SECONDS),
                stop=stop_after_attempt(settings.MAX_RETRIES),
            )
            send_email_retryer(_perform_send_email)
    except Exception as e:
        logger.info(
            f'ERROR : Unable to send email {instance.pk} '
            f'to {instance.receiver.email}, {str(e)}'
        )
        instance.resource_status = 'send-error'
        instance.resource_message = 'Unable to send email'
        save_instance_if_async(instance=instance, is_async=is_async)



# coding: utf-8
import pendulum
from django.utils.translation import gettext_lazy as _
from django.conf import settings

from rest_framework import authentication
from rest_framework import exceptions

from concrete_datastore.concrete.models import (  # pylint:disable=E0611
    SecureConnectToken,
    SecureConnectCode,
    AuthToken,
    Group,
)


def default_mfa_rule(user):
    return settings.USE_TWO_FACTOR_AUTH and user.is_at_least_admin


def default_backend_group_creation_rule(user, groups):
    for group_name in groups:
        group, created = Group.objects.get_or_create(name=group_name)
        user.concrete_groups.add(group)


def expire_temporary_tokens(user):
    # TODO: Get all temporary tokens from user and expire those out of date
    temp_token_expiry_seconds = getattr(
        settings, 'TWO_FACTOR_CODE_TIMEOUT_SECONDS', 600
    )
    now = pendulum.now('utc')
    expiry_token_limit = now.subtract(seconds=temp_token_expiry_seconds)
    user.temporary_tokens.filter(
        expired=False, creation_date__lte=expiry_token_limit
    ).update(expired=True)


def api_token_has_expired(token):
    if getattr(token, 'expired', False) is True:
        token.delete()
        return True
    token_can_expire = settings.API_TOKEN_EXPIRY > 0
    extra_period = settings.EXPIRY_EXTRA_PERIOD
    if token_can_expire:
        now = pendulum.now('utc')
        expiration_date = pendulum.instance(token.expiration_date)
        expiry_period_over = now > expiration_date
        expiry_spare_period_over = (
            extra_period == 0
            or now.diff(token.last_action_date).in_minutes() > extra_period
        )
        token_expired = expiry_period_over and expiry_spare_period_over
        if token_expired:
            token.delete()
            return True
    return False


def ensure_secure_connect_instance_is_not_expired(
    instance, expiration_limit_in_seconds
):
    # if the instance is not expired, return True
    # false otherwise
    if expiration_limit_in_seconds:
        now = pendulum.now('utc')
        secure_connect_instance_expired = (
            now.diff(instance.creation_date).in_seconds()
            >= expiration_limit_in_seconds
        )
        if secure_connect_instance_expired:
            instance.expired = True
            instance.save()
            return False
    return True


class TokenExpiryAuthentication(authentication.TokenAuthentication):
    def authenticate_credentials(self, key):
        model = AuthToken
        try:
            token = model.objects.select_related('user').get(key=key)
        except model.DoesNotExist:
            raise exceptions.AuthenticationFailed(_('Invalid token.'))

        if not token.user.is_active:
            raise exceptions.AuthenticationFailed(
                _('User inactive or deleted.')
            )

        if api_token_has_expired(token):
            raise exceptions.AuthenticationFailed(_('Token expired'))
        # Check if the secure token is expired and expire in database
        for secure_token in SecureConnectToken.objects.filter(
            user=token.user, expired=False
        ):
            ensure_secure_connect_instance_is_not_expired(
                secure_token, settings.SECURE_CONNECT_TOKEN_EXPIRY_TIME_SECONDS
            )
        # Check if the secure token is expired and expire in database
        for secure_code in SecureConnectCode.objects.filter(
            user=token.user, expired=False
        ):
            ensure_secure_connect_instance_is_not_expired(
                secure_code, settings.SECURE_CONNECT_CODE_EXPIRY_TIME_SECONDS
            )

        return (token.user, token)


class URLTokenExpiryAuthentication(TokenExpiryAuthentication):
    """
    This class allow a user to authenticate using the query param
    c_auth_with_token in the URL i.e. by appending ?c_auth_with_token=<value>
    """

    def authenticate(self, request):
        token = request.GET.get('c_auth_with_token', '')
        if not token:
            return

        if len(token) != 40:
            msg = _('Invalid token : {}'.format(repr(token)))
            raise exceptions.AuthenticationFailed(msg)

        return self.authenticate_credentials(token)



# coding: utf-8
from __future__ import unicode_literals, absolute_import, print_function
import json


def parse_json(fd):
    data = json.load(fd)
    return data


def parse_json_file(path):
    with open(path, 'r') as fd:
        return parse_json(fd)



# coding: utf-8
class DatamodelYamlToHtml:
    """
    This class renders a json into a yaml format for html input.
    A JSON content should be given to this class.
    The return value of the "render_yaml()" method can be used in a HTML
    template as follows:
        <pre>{{ rendered_result | safe }}</pre>
    """

    bool_color = "#bc3753"
    string_color = "#0a6969"

    def __init__(self, content, indent=4):
        self.indent = indent
        self.json_content = content

    def format_with_color(self, val, color=None):
        if not val:
            val = '""'
        if color is not None:
            return f'<span style="color:{color}">{val}</span>'
        return f'<span>{val}</span>'

    def _render_yaml(self, content, indent_counter, is_list=False):
        if isinstance(content, bool):
            str_value = str(content).lower()
            v = self.format_with_color(val=str_value, color=self.bool_color)
            if is_list is True:
                v = f'{" " * (indent_counter) * self.indent}- {v}'
            return f'{v}\n'
        if content is None:
            v = self.format_with_color(val='null', color=self.bool_color)
            if is_list is True:
                v = f'{" " * (indent_counter) * self.indent}- {v}'
            return f'{v}\n'
        if isinstance(content, list):
            if len(content) == 0:
                return self.format_with_color("[]\n")
            return_value = ''.join(
                [
                    self._render_yaml(
                        elt, indent_counter=indent_counter, is_list=True
                    )
                    for elt in content
                ]
            )
            return f'\n{return_value}'
        if isinstance(content, dict):
            if not content:
                return self.format_with_color("{}\n")
            v = "\n"
            for key, value in content.items():
                formatted_key = self.format_with_color(key)
                add_list_counter = 0
                if is_list:
                    add_list_counter = 1
                value_indent_counter = indent_counter + add_list_counter + 1

                formatted_value = self._render_yaml(
                    content=value, indent_counter=value_indent_counter
                )
                if is_list:
                    v = f'{" " * (indent_counter) * self.indent}- {formatted_key}: {formatted_value}'
                    is_list = False
                    indent_counter += 1
                else:
                    v = f'{v}{" " * indent_counter * self.indent}{formatted_key}: {formatted_value}'
            return v

        v = self.format_with_color(val=content, color=self.string_color)
        if is_list is True:
            v = f'{" " * (indent_counter) * self.indent}- {v}'
        return f'{v}\n'

    def render_yaml(self):
        return self._render_yaml(content=self.json_content, indent_counter=0)
